@@ -6,14 +6,15 @@ deliver the services provided by Raindrop.IO as a set of class-based methods on
 For example, to create a new raindrop, use Raindrop.create_link(...); a collection would be Collection.create(...) etc.
 
 """
+
 from __future__ import annotations
 
 import enum
-import json
 from datetime import datetime
 from pathlib import Path
-from typing import Any, Optional
+from typing import Any
 
+import requests
 from pydantic import (
     BaseModel,
     EmailStr,
@@ -25,7 +26,7 @@ from pydantic import (
     validator,
 )
 
-from .api import tAPI  # ie. for typing only...
+from .api import T_API  # ie. for typing only...
 
 __all__ = [
     "Access",
@@ -48,6 +49,7 @@ __all__ = [
 # Base URL for Raindrop IO's API
 URL = "https://api.raindrop.io/rest/v1/{path}"
 
+
 ################################################################################
 # Utility methods
 ################################################################################
@@ -56,13 +58,27 @@ def _collect_other_attributes(cls, v):
     skip_attrs = "_id"  # We don't need to store alias attributes again (pydantic will take care of)
     v["other"] = dict()
     for attr, value in v.items():
-        if attr not in cls.__fields__:
-            if attr not in skip_attrs:
-                if value:
-                    v["other"][attr] = value
+        if value and attr not in cls.__fields__ and attr not in skip_attrs:
+            v["other"][attr] = value
     return v
 
 
+def _resolve_parent_reference(parent_reference: dict | int | None) -> int | None:
+    """Convert a Raindrop.IO parent reference dict to just the respective ID of the parent collection.
+
+    For a child collection that has a parent, the reference to the parent received from Raindrop.IO is:
+
+    {"$id": 12345678, "$ref": 'collections'}.
+
+    We don't need the $ref part (at least I don't believe so), so simply pull the $id key.
+    """
+    if parent_reference is None:
+        return None
+    elif isinstance(parent_reference, int):
+        return parent_reference
+    return parent_reference.get("$id")
+
+
 ################################################################################
 # Enumerated types
 ################################################################################
@@ -152,7 +168,9 @@ class CollectionRef(BaseModel):
 
 
 # We define the 3 "system" collections in the Raindrop environment:
-CollectionRef.All = CollectionRef(**{"$id": 0})
+CollectionRef.All = CollectionRef(
+    **{"$id": 0},
+)  # Note: "all" here does NOT include Trash.
 CollectionRef.Trash = CollectionRef(**{"$id": -99})
 CollectionRef.Unsorted = CollectionRef(**{"$id": -1})
 
@@ -185,16 +203,16 @@ class Collection(BaseModel):
         cover: URL of the collection's cover.
         created: When the collection was created.
         expanded: Whether the collection's sub-collection are expanded (on the interface)
-        lastUpdate: When the collection was last updated.
+        last_update: When the collection was last updated.
         parent: Parent ID of this is a sub-collection.
         public: Are contents of this collection available to non-authenticated users?
-        sort: The order of the collection. Defines the position of the collection all other
-          collections at the same level in the tree.
+        sort: The order of the collection. Defines the position of the collection against
+          all other collections at the same level in the tree (only used for sub-collections?)
         view: Current view style of the collection, e.g. list, simple, grid etc.
-        other: All other attributes received from Raindrop's API.
+        other: All other attributes received from Raindrop's API (see Warning below)
 
     Warning:
-        Attributes in `other` are NOT OFFICIALLY SUPPORTED!.
+        Attributes in `other` are *NOT* OFFICIALLY SUPPORTED...use at your own risk!
     """
 
     id: int = Field(None, alias="_id")
@@ -208,23 +226,29 @@ class Collection(BaseModel):
     cover: list[str] | None = Field(default_factory=list)
     created: datetime | None
     expanded: bool = False
-    lastUpdate: datetime | None
-    parent: int | None  # Id of parent collection
+    last_update: datetime | None
+    parent: int | None  # Id of parent collection (if any)
     public: bool | None
-    sort: NonNegativeInt | None
+    sort: int | None
     view: View | None
 
     # Per API Doc: "Our API response could contain other fields, not described above.
     # It's unsafe to use them in your integration! They could be removed or renamed at any time."
     other: dict[str, Any] = {}
 
+    # Used to convert parent reference's of sub-collections to simply id's of the respective parent collection.
+    _extract_parent_id = validator("parent", pre=True, allow_reuse=True)(
+        _resolve_parent_reference,
+    )
+
     @root_validator(pre=True)
-    def _validator_other_attributes(cls, v):
+    # FIXME: noqa here is because work-around in https://github.com/pydantic/pydantic/issues/568 doesn't work!
+    def _validator(cls, v):  # noqa: N805
         """Gather all non-recognised/unofficial attributes into a single attribute."""
         return _collect_other_attributes(cls, v)
 
     @classmethod
-    def get_root_collections(cls, api: tAPI) -> list[Collection]:
+    def get_root_collections(cls, api: T_API) -> list[Collection]:
         """Get **root** Raindrop collections.
 
         Args:
@@ -236,7 +260,7 @@ class Collection(BaseModel):
         Note:
             Since Raindrop allows for collections to be nested, the RaindropIO's API distinguishes between Collections
             at the top-level/root of a collection hierarchy versus those all that are below the top level, aka 'child'
-            collections. Thus, use ``get_root_collections`` to get all Collections without parents and
+            or 'sub' collections. Thus, use ``get_root_collections`` to get all Collections without parents and
             ``get_child_collections`` for all Collections with parents.
         """
         ret = api.get(URL.format(path="collections"))
@@ -244,7 +268,7 @@ class Collection(BaseModel):
         return [cls(**item) for item in items]
 
     @classmethod
-    def get_child_collections(cls, api: tAPI) -> list[Collection]:
+    def get_child_collections(cls, api: T_API) -> list[Collection]:
         """Get the **child** Raindrop collections (ie. all below root level).
 
         Args:
@@ -264,7 +288,7 @@ class Collection(BaseModel):
         return [cls(**item) for item in items]
 
     @classmethod
-    def get_collections(cls, api: tAPI) -> list[Collection]:
+    def get_collections(cls, api: T_API) -> list[Collection]:
         """Query for all non-system collections (essentially a convenience wrapper, combining root & child Collections).
 
         Args:
@@ -277,7 +301,7 @@ class Collection(BaseModel):
         return cls.get_root_collections(api) + cls.get_child_collections(api)
 
     @classmethod
-    def get(cls, api: tAPI, id: int) -> Collection:
+    def get(cls, api: T_API, id: int) -> Collection:
         """Return a Raindrop Collection instance based on it's id.
 
         Args:
@@ -298,7 +322,7 @@ class Collection(BaseModel):
     @classmethod
     def create(
         cls,
-        api: tAPI,
+        api: T_API,
         title: str,
         cover: list[str] | None = None,
         expanded: bool | None = None,
@@ -352,7 +376,7 @@ class Collection(BaseModel):
     @classmethod
     def update(
         cls,
-        api: tAPI,
+        api: T_API,
         id: int,
         cover: list[str] | None = None,
         expanded: bool | None = None,
@@ -395,7 +419,7 @@ class Collection(BaseModel):
         return cls(**item)
 
     @classmethod
-    def delete(cls, api: tAPI, id: int) -> None:
+    def delete(cls, api: T_API, id: int) -> None:
         """Delete a Raindrop collection.
 
         Args:
@@ -409,7 +433,7 @@ class Collection(BaseModel):
         api.delete(URL.format(path=f"collection/{id}"), json={})
 
     @classmethod
-    def get_or_create(cls, api: tAPI, title: str) -> Collection:
+    def get_or_create(cls, api: T_API, title: str) -> Collection:
         """Get a Raindrop collection based on it's **title**, if it doesn't exist, create it.
 
         Args:
@@ -458,12 +482,12 @@ class UserConfig(BaseModel):
     other: dict[str, Any] = {}
 
     @validator("last_collection", pre=True)
-    def cast_last_collection_to_ref(cls, v):
+    def cast_last_collection_to_ref(cls, v):  # noqa: N805
         """Cast last_collection provided as a raw int to a valid CollectionRef."""
         return CollectionRef(**{"$id": v})
 
     @root_validator(pre=True)
-    def _validator_other_attributes(cls, v):
+    def _validator_other_attributes(cls, v):  # noqa: N805
         """Gather all non-recognised/unofficial attributes into a single attribute."""
         return _collect_other_attributes(cls, v)
 
@@ -473,7 +497,7 @@ class UserFiles(BaseModel):
 
     used: int
     size: PositiveInt
-    lastCheckPoint: datetime
+    last_checkpoint: datetime = Field(None, alias="lastCheckpoint")
 
 
 class User(BaseModel):
@@ -481,18 +505,18 @@ class User(BaseModel):
 
     id: int = Field(None, alias="_id")
     email: EmailStr
-    email_MD5: str | None = None
+    email_md5: str | None = Field(None, alias="email_MD5")
     files: UserFiles
-    fullName: str
+    full_name: str = Field(None, alias="fullName")
     groups: list[Group]
     password: bool
     pro: bool
-    proExpire: datetime | None = None
+    pro_expire: datetime | None = Field(None, alias="proExpire")
     registered: datetime
     config: UserConfig
 
     @classmethod
-    def get(cls, api: tAPI) -> User:
+    def get(cls, api: T_API) -> User:
         """Get all the information about the Raindrop user associated with the API token."""
         user = api.get(URL.format(path="user")).json()["user"]
         return cls(**user)
@@ -509,7 +533,7 @@ class SystemCollection(BaseModel):
         - The *Trash* collection contains Raindrops that have been recently deleted.
 
     You won't use this class directly on behalf of individual Raindrops, rather, its definition is on behalf of
-    a simple "status" call available from the Raindrop.io API.
+    a small set of simple "status" calls available from the Raindrop.io API, specifically `get_counts` and `get_meta`.
     """
 
     id: int = Field(None, alias="_id")
@@ -517,7 +541,7 @@ class SystemCollection(BaseModel):
     title: str | None
 
     @root_validator(pre=False)
-    def _map_systemcollection_id_to_title(cls, values):
+    def _validator(cls, values):  # noqa: N805
         """Map the hard-coded id's of the System Collections to the descriptions used on the UI."""
         _titles = {
             CollectionRef.Unsorted.id: "Unsorted",
@@ -528,11 +552,25 @@ class SystemCollection(BaseModel):
         return values
 
     @classmethod
-    def get_status(cls, api: tAPI) -> User:
-        """Get the count of Raindrops across all 3 *system* collections."""
+    def get_counts(cls, api: T_API) -> list[Collection]:
+        """Get the count of Raindrops in each of the 3 *system* collections."""
         items = api.get(URL.format(path="user/stats")).json()["items"]
         return [cls(**item) for item in items]
 
+    @classmethod
+    def get_meta(cls, api: T_API) -> dict:
+        """Get the 'meta' slug from the root/system Collection.
+
+        Contains information about:
+
+        - Last date/time any bookmark was changed.
+
+        - The number of broken links in bookmarks.
+
+        - If your account is a "pro" level.
+        """
+        return api.get(URL.format(path="user/stats")).json()["meta"]
+
 
 class File(BaseModel):
     """Represents the attributes associated with a file within a document-based Raindrop."""
@@ -545,8 +583,10 @@ class File(BaseModel):
 class Cache(BaseModel):
     """Represents the cache information of Raindrop."""
 
+    # Per issue #5, we can't rely on Raindrop to always return a non-zero value for `size`, thus
+    # instead of `PositiveInt`, we use `int`.
     status: CacheStatus
-    size: PositiveInt | None = None
+    size: int | None = None
     created: datetime | None = None
 
 
@@ -568,7 +608,7 @@ class Raindrop(BaseModel):
         domain: Hostname of a link, ie. if a Raindrop has link: `https://www.google.com?search=SomeThing`,
           domain is `www.google.com`.
         excerpt: Description associated with this Raindrop (maximum length: 10k!)
-        lastUpdate: When this Raindrop was last updated.
+        last_update: When this Raindrop was last updated.
         link: For a link-based Raindrop, the full URL.
         media: Covers list.
         tags: A list of Tags associated with the Raindrop.
@@ -592,7 +632,8 @@ class Raindrop(BaseModel):
     created: datetime | None
     domain: str | None
     excerpt: str | None  # aka 'Description' on the Raindrop UI.
-    lastUpdate: datetime | None
+    file: File | None
+    last_update: datetime | None = Field(None, alias="lastUpdate")
     link: HttpUrl | None
     media: list[dict[str, Any]] | None
     tags: list[str] | None
@@ -603,7 +644,6 @@ class Raindrop(BaseModel):
     # "Other" fields:
     broken: bool | None
     cache: Cache | None
-    file: File | None
     important: bool | None  # aka marked as Favorite.
 
     # Per API Doc: "Our API response could contain other fields, not described above.
@@ -611,28 +651,36 @@ class Raindrop(BaseModel):
     other: dict[str, Any] = {}
 
     @root_validator(pre=True)
-    def _validator_other_attributes(cls, v):
+    def _validator(cls, v):  # noqa: N805
         """Gather all non-recognised/unofficial attributes into a single attribute."""
         return _collect_other_attributes(cls, v)
 
     @classmethod
-    def get(cls, api: tAPI, id: int) -> Raindrop:
+    def get(cls, api: T_API, id: int) -> Raindrop:
         """Return a Raindrop bookmark based on it's id."""
         item = api.get(URL.format(path=f"{id}")).json()["item"]
         return cls(**item)
 
+    @classmethod
+    def cache(cls, api: T_API, id: int) -> requests.Response:
+        """Return the requests on behalf of a permanent copy of the specified Raindrop."""
+        # Note: In testing in 2024-01, while I was able to get a URL back in this response
+        # (after a 307 redirect), the URL did NOT work against S3...(essentially an "item not
+        # found" from the S3 infrastructure).
+        return api.get(URL.format(path=f"raindrop/{id}/cache"))
+
     @classmethod
     def create_link(
         cls,
-        api: tAPI,
+        api: T_API,
         link: str,
-        collection: Optional[Collection | CollectionRef, int] = None,
+        collection: (Collection | CollectionRef, int) | None = None,
         cover: str | None = None,
         excerpt: str | None = None,
         important: bool | None = None,
         media: list[dict[str, Any]] | None = None,
         order: int | None = None,
-        pleaseParse: bool = False,  # If set, asks API to automatically parse metadata in the background
+        please_parse: bool = False,  # If set, asks API to automatically parse metadata in the background
         tags: list[str] | None = None,
         title: str | None = None,
     ) -> Raindrop:
@@ -657,7 +705,7 @@ class Raindrop(BaseModel):
 
             order: Optional, Order of Raindrop in respective collection, ie. set to 0 to make Raindrop first.
 
-            pleaseParse: Optional, Flag that asks API to automatically parse metadata in the background
+            please_parse: Optional, Flag that asks API to automatically parse metadata in the background
               (not exactly sure which this implies, message me if you know! ;-)
 
             tags: Optional, List of tags to associated with this Raindrop.
@@ -668,41 +716,15 @@ class Raindrop(BaseModel):
             ``Raindrop`` instance created.
 
         Note:
-            We don't allow you to set either ``created`` or ``lastUpdate`` attributes. They will be set appropriately
-            by the RaindropIO service on your behalf.
+            We don't allow you to set either ``created`` or ``last_update`` attributes. They will be
+            set appropriately by the RaindropIO service on your behalf.
         """
-        # NOTE: We have small code style conflict here between Vulture
-        # and Ruff, specifically:
-
-        # Vulture will report all the optional variables above as
-        # "unused". This is clearly technically true as their only
-        # appearance below is as a string to the arg below from which
-        # we take the actual value of the inbound argument from
-        # locals(), instead of referring to the argument explicitly.
-        #
-        # However, converting all of these to a (rather lengthy set)
-        # of simple statements like:
-        #
-        # if <arg> is not None:
-        #     args[<arg>] = <arg>
-        #
-        # keeps Vulture happy but now trips the McCabe complexity
-        # metric(!) implemented in our Ruff pre-commit pass due to the
-        # number of conditionals in a single method!
-        #
-        # While, we /could/, convert to kwargs, we lose the documentation
-        # and verbosity available with explicit arguments.
-        #
-        # Thus, for now, we **leave** vulture out of pre-commit and
-        # simply run it manually as necessary (through our justfile)
-
-        # Setup the args that will be passed to the underlying
-        # Raindrop API, only link is absolutely required, rest are
-        # optional!
+        # Setup the args that will be passed to the underlying Raindrop API, only link is
+        # absolutely required, rest are optional!
         args: dict[str, Any] = dict(type=RaindropType.link, link=link)
 
-        if pleaseParse:
-            args["pleaseParse"] = {}
+        if please_parse:
+            args["please_parse"] = {}
 
         for attr in [
             "cover",
@@ -719,7 +741,7 @@ class Raindrop(BaseModel):
         if collection is not None:
             # <collection> arg could be **either** an actual collection
             # or simply an int collection "id" already, handle either:
-            if isinstance(collection, (Collection, CollectionRef)):
+            if isinstance(collection, Collection | CollectionRef):
                 args["collection"] = {"$id": collection.id}
             else:
                 args["collection"] = {"$id": collection}
@@ -730,10 +752,10 @@ class Raindrop(BaseModel):
     @classmethod
     def create_file(
         cls,
-        api: tAPI,
+        api: T_API,
         path: Path,
         content_type: str,
-        collection: Optional[Collection | CollectionRef, int] = CollectionRef.Unsorted,
+        collection: (Collection | CollectionRef, int) | None = CollectionRef.Unsorted,
         tags: list[str] | None = None,
         title: str | None = None,
     ) -> Raindrop:
@@ -773,19 +795,19 @@ class Raindrop(BaseModel):
 
         # NOTE: "put_file" arguments and structure here confirmed through communication
         #       with RustemM on 2022-11-29 and his subsequent update to API docs.
-        if isinstance(collection, (Collection, CollectionRef)):
+        if isinstance(collection, Collection | CollectionRef):
             data = {"collectionId": str(collection.id)}
         else:
             data = {"collectionId": str(collection)}
 
-        files = {"file": (path.name, open(path, "rb"), content_type)}
-        item = api.put_file(url, path, data, files).json()["item"]
-        raindrop = cls(**item)
+        with open(path, "rb") as fh_:
+            files = {"file": (path.name, fh_, content_type)}
+            item = api.put_file(url, path, data, files).json()["item"]
+            raindrop = cls(**item)
 
-        # The Raindrop API's "Create Raindrop From File" does not allow us to set other attributes, thus,
-        # we need to check if any of the possible attributes need to be set and do so explicitly with another
-        # call to "update" the Raindrop we just
-        # created.:
+        # Raindrop's "Create Raindrop From File" does not allow us to set other attributes,
+        # thus, we need to check if any of the possible attributes need to be set and do so
+        # explicitly with another call to "update" the Raindrop we just created.
         args: dict[str, Any] = {}
         if title is not None:
             args["title"] = title
@@ -801,16 +823,16 @@ class Raindrop(BaseModel):
     @classmethod
     def update(
         cls,
-        api: tAPI,
+        api: T_API,
         id: int,
-        collection: Optional[Collection | CollectionRef, int] = None,
+        collection: (Collection | CollectionRef, int) | None = None,
         cover: str | None = None,
         excerpt: str | None = None,
         important: bool | None = None,
         link: str | None = None,
         media: list[dict[str, Any]] | None = None,
         order: int | None = None,
-        pleaseParse: bool | None = False,
+        please_parse: bool | None = False,
         tags: list[str] | None = None,
         title: str | None = None,
     ) -> Raindrop:
@@ -836,7 +858,7 @@ class Raindrop(BaseModel):
 
             order: Optional, Change order of Raindrop in respective collection.
 
-            pleaseParse: Optional, Flag that asks API to automatically parse metadata in the background
+            please_parse: Optional, Flag that asks API to automatically parse metadata in the background
                 (not exactly sure which this implies, message me if you know! ;-)
 
             tags: Optional, New list of tags to associate with this Raindrop.
@@ -849,8 +871,8 @@ class Raindrop(BaseModel):
         # Setup the args that will be passed to the underlying Raindrop API
         args: dict[str, Any] = {}
 
-        if pleaseParse:
-            args["pleaseParse"] = {}
+        if please_parse:
+            args["please_parse"] = {}
 
         for attr in [
             "cover",
@@ -868,7 +890,7 @@ class Raindrop(BaseModel):
         if collection is not None:
             # <collection> arg could be **either** an actual collection
             # or simply an int collection "id" already, handle either:
-            if isinstance(collection, (Collection, CollectionRef)):
+            if isinstance(collection, Collection | CollectionRef):
                 args["collection"] = collection.id
             else:
                 args["collection"] = collection
@@ -878,7 +900,7 @@ class Raindrop(BaseModel):
         return cls(**item)
 
     @classmethod
-    def delete(cls, api: tAPI, id: int) -> None:
+    def delete(cls, api: T_API, id: int) -> None:
         """Delete a Raindrop bookmark.
 
         Args:
@@ -894,28 +916,21 @@ class Raindrop(BaseModel):
     @classmethod
     def _search_paged(
         cls,
-        api: tAPI,
+        api: T_API,
         collection: CollectionRef = CollectionRef.All,
+        search: str | None = None,
         page: int = 0,
         perpage: int = 50,
-        word: str | None = None,
-        tag: str | None = None,
-        important: bool | None = None,
     ) -> list[Raindrop]:
         """Lower-level search for bookmarks on a "paged" basis.
 
-        Raindrop's search API works on a "paged" basis. This method implements
-        the underlying search reflecting paging. While the ``search`` method hides it completely.
+        Raindrop's search API works on a "paged" basis. This method implements the underlying
+        search reflecting paging (while the primary ``search`` method below hides it
+        completely).
         """
-        args: list[dict[str, Any]] = []
-        if word is not None:
-            args.append({"key": "word", "val": word})
-        if tag is not None:
-            args.append({"key": "tag", "val": tag})
-        if important is not None:
-            args.append({"key": "important", "val": important})
-
-        params = {"search": json.dumps(args), "perpage": perpage, "page": page}
+        params = {"perpage": perpage, "page": page}
+        if search:
+            params["search"] = search
         url = URL.format(path=f"raindrops/{collection.id}")
         results = api.get(url, params=params).json()
         return [cls(**item) for item in results["items"]]
@@ -923,11 +938,9 @@ class Raindrop(BaseModel):
     @classmethod
     def search(
         cls,
-        api: tAPI,
+        api: T_API,
         collection: Collection | CollectionRef = CollectionRef.All,
-        word: str | None = None,
-        tag: str | None = None,
-        important: bool | None = None,
+        search: str | None = None,
     ) -> list[Raindrop]:
         """Search for Raindrops.
 
@@ -937,13 +950,9 @@ class Raindrop(BaseModel):
             collection: Optional, ``Collection`` (or ``CollectionRef``) to search over.
                 Defaults to ``CollectionRef.All``.
 
-            word: Optional, search string to search Raindrops for (see
+            search: Optional, search string to search Raindrops for (see
                 `Raindrop.io Search Help <https://help.raindrop.io/using-search#operators>`_ for more information.
 
-            tag: Optional, tag name to search for.
-
-            important: Optional, flag indicating if only important/favorite Raindrops should be search over (or not).
-
         Returns:
             A (potentially empty) list of Raindrops that match the search criteria provided.
         """
@@ -953,9 +962,7 @@ class Raindrop(BaseModel):
             api,
             collection,
             page=page,
-            word=word,
-            tag=tag,
-            important=important,
+            search=search,
         ):
             results.extend(raindrops)
             page += 1
@@ -969,7 +976,7 @@ class Tag(BaseModel):
     count: int
 
     @classmethod
-    def get(cls, api: tAPI, collection_id: int = None) -> list[Tag]:
+    def get(cls, api: T_API, collection_id: int | None = None) -> list[Tag]:
         """Get all the tags currently defined, either in a specific collections or across all collections.
 
         Args:
@@ -987,7 +994,7 @@ class Tag(BaseModel):
         return [Tag(**item) for item in items]
 
     @classmethod
-    def delete(cls, api: tAPI, tags: list[str]) -> None:
+    def delete(cls, api: T_API, tags: list[str]) -> None:
         """Delete one or more Tags.
 
         Args:
